The Scan Operator (SO) is an integration tool for scan data and DCS data, which is meant to be used during the QC procedure of the Rd53a modules. It uses multiple localDB scripts as well as some integrated tools to achieve these features:

  - Calls Yarr repeatedly to **run a sequence of scans on Rd53a**.
  - Gets the config files from Prod. DB if the module is registered there; creates them offline otherwise.
  - **Monitors the Power Supply** (PS) powering the ASIC and storing the DCS data in **InfluxDB**
  - Synchronizes the DCS data and the scan data in **LocalDB**.
  - Makes very easy to edit chip-by-chip config parameters (trims, chip name, rx and tx, ...)

All the DCS-related functionality is optional. The SO provides a script based on labRemote (LR) to monitor DCS in real time, but any other external tool can also be used as long as it is capable of uploading the DCS data to InfluxDB.

# Quick Tutorial

## 1. Get the code
If you want to use the SO's DCS integrated tool, start by cloning and building labRemote. *You don't need to do it in other case*.  

Make sure to enable the Python bindings using cmake's `-DUSE_PYTHON=on` flag. For a detailed installarion guide, follow [labRemote's readme](https://gitlab.cern.ch/berkeleylab/labRemote/).

```bash
git clone https://gitlab.cern.ch/berkeleylab/labRemote.git
cd labRemote
git submodule update --init --recursive

mkdir build && cd build
yum install python3-devel
cmake3 -DUSE_PYTHON=on .. && make -j6
```

labRemote's commit [24e7a55b](https://gitlab.cern.ch/berkeleylab/labRemote/-/commit/24e7a55b5e943db7fb6c6252c610b40d9eb8cf62) has been confirmed to work with no problems together with the Scan Operator. We recommend using the latest version and switching to this one only in case of LR-related errors.

**Clone also somewhere the Scan Operator**
```bash
git clone --recursive https://gitlab.cern.ch/YARR/utilities/scan-operator.git
```

**Use the appropiate version of Yarr**
The latest we have tested is [680f0adc](https://gitlab.cern.ch/YARR/YARR/-/commit/680f0adc7d91c611e43039835f92eae7c50da830) on branch "`devel-localdb`", but more recent versions will most likely also work fine. We recommend starting from the latest one.

## 2. Basic configuration
The Scan Operator uses up to four configuration files, depending on the functions you want to make use of. All of them are placed under `configs/`, and they can be renamed or moved anywhere else. The following is the only required one.

### General configuration file

Under `scan_operator/configs/`, this [`config.json`](configs/config.json) file contains YARR's basic configuration, module information and some general dcs-related settings.

Most of the items are self-explanatory. Here are the ones that may require some indications to fill.

  - **`.YARR.scan_list`**: Place here the sequence of scans that you want to run. For a threshold / tot tuning, the second argument in the bracket represents the target threshold / tot (e.g. `["diff_retune_pixelthreshold", 1000]` will retune the pixel threshold of the differential FE to 1000 e), whilst `["diff_tune_globalpreamp", 7]` will tune the ToT to 7). The target charge for ToT tuning / scan is specified under `.YARR.scan` and it's assumed to be the same for all the ToT-related scans.  

  - **`.modules`**: Write here the configs for as many modules as you want. The file comes with an example for a SCC and an example for a quad. You will choose between the modules you have defined here in the command line. Notes:
    - The name of each module (e.g. `moduleid_quad`) should match the Serial Number of the module.
    - For each chip:
      - For any key written in `"GlobalConfig"`, the Scan Operator will:
        - Look for it in Yarr's general chip config file (e.g. `rd53a_test.json`) under `.RD53A.GlobalConfig`. 
        - If found, change the value to the one specified by the user.
        - If not found, create it and continue.
      - The same goes for the `"connectivity"` and the `"Parameter"` section.  
  
  - **`.dcs_control`**: *Fill this only if you are planning to **use the SO to** monitor your PS while running the scans.*
    - under `.dcs_control.channels`, write one entry for each channel you want to monitor (`name`, `voltage` and `current`). You will link them to your hardware later.
      - The `voltage` and `current` keys will be ignored if you don't use "`-o`" in the command line (see [below](#dcs-related)). 
      - Different channels have to have different names.


### Powersupply  configuration file

*Skip this section if you don't want to use the SO to monitor DCS data.*

[The `powersupply.json`](configs/powersupply.json) config file contains the actual information to establish communication with the PS. These are some fields you may want to change:

- **`.devices`**: Write here the PS / PSs you want to use. A list of supported "`hw-model`" and "`protocol`" is available in [labRemote's README.md](https://gitlab.cern.ch/berkeleylab/labRemote#python-bindings). To find the port your PS is connected to, look at the output of `ls -l /dev/serial/by-id/`.
  
- **`.channels`**: You can have here as many channels as you want, but only the ones appearing in `config.json` will be used. Under `.channels`, `"device"` should be one of the ones in the upper section, and `"channel"` is the actual output channel number of the PS (e.g. from 1 to 4 for the R&S HMP4040).
  


### InfluxDB connectivity configuration file


*Skip this section if you don't want to use the SO to monitor DCS data **and** you don't want to synchronize DCS and scan data in localDB*.

The file `configs/influxdb_connectivity.json` contains the basic information needed to stablish connection with InfluxDB. Also, it combines various configuration files localDB needs to retrieve DCS data from InfluxDB.

Setting up this file is easy. These are the fields that need a short explanation:

  - `.influxdb_cfg`: If you are planning on monitoring the DCS, but you are not interested in synchronizing the data with any scan in localDB, that's the only section you should worry about in this file.
  
  - `.influxdb_cfg.measurement`: The measurement (data table) where to upload data to in InfluxDB.  
  

  - `.environment`: List here the title of every set of data you want to upload to localDB, *as you want it to appear in localDB* (e.g. vdda, temperature, HV, ...)  


  - `.environments.measurement`: The measurement where to read data from in InfluxDB

  - `.environments.dcslist[i].key`: Which key you want your data to be associated to (in localDB); has to match one of the ones declared in `.environment`
  
  - `.environments.dcslist[i].data_name`: The title of the column **in InfluxDB** where to read data from, to be associated with `.environments.dcslist[i].key` **in localDB**. If you are using the SO (and not your own DCS tool) to monitor your PS, `data_name` is not a free parameter. It is named as `AX_C`, where:
    - `A`: `meas` for the measured values and `sett` for the setting values
    - `X`: `V` or `I`
    - `C`: PS Channel, as you named it in the main config file. 
    E.g. `measV_LV` will contain the sense values of V read on the channel named "LV".


[//]: # (### Configuration file for IV / VI scans)


[//]: # "This `ivscan.json` file is used by `libDCS/ivscan.py`, a script based on labRemote to perform (sensor) IV and (module) VI scans and output the data in a format readable suitable to be uploaded to localBD (to be decided). It has no direct relation to the Scan Operator."

[//]: # (It contains one section for sensor IV and another one for module VI, following both of them the same scheme:)

[//]: # (`.[what].target_channel`: The channel where to run the scan)
[//]: # (`.[what].target_channel.measurements_per_step`: The `ivscan.py` can take more than one measurement per step, so it calculates the mean and the std to give a more accurate result.)
[//]: # (`.[what].target_channel.sleeping_time_s`: The sleeping time between step, in seconds)



## 3. Run the Scan Operator

```
$ ./ScanOperator.sh
Help:
  -j [str, path] : path to json configuration file
  -m [str]       : Serial number of the module
  -W             : upload Yarr's results to localDB (scanConsole's -W)
  -Q             : QC mode (upload / retrieve from iTk PD, etc) (scanConsole's -Q).
  -o             : turn on / off the output of the PS before and after all the scans
  -d [str, path] : powersupply.json. Monitor the PS and upload the DCS data to influxDB.
  -t [str, path] : influxdb_connectivity.json. Move DCS data from InfluxDB to LocalDB after each scan.
  -c             : create / override Yarr's config files for your chip based on config.json.
  -q [int]       : "quiet" mode (cleaner output), 0, or 1. Default 0
```

A good starting point is following these [example use cases](#example-calls).


### List of arguments

#### Required

- `-j:` Path to your `config.json`
- `-m:` Serial Number of the module. Chose from the ones you have defined in `config.json`

#### Optional

- `-c:` Create Yarr config files. Necessary files such as the `controller` or `connectivity` **config files will be created or overwritten** if `-c` is specified. Use this option after making any change in `config.json`, or if you are using the software for the first time. All the files are created in the Scan Operator folder, so everything you have under Yarr's one remains untouched. Summarizing, "-c" does the following.
  - Retrieves the files from localDB if the module is already registered and the connection is good. Creates them offline otherwise.
  - Edits them based on `config.json`
  - if -W is used, reuploads them to localDB, associated to every scan.  
  
- `-W:` This one is equivalent to the `-W` option used when calling Yarr's ScanConsole  
- `-Q:` This one is equivalent to the `-Q` option you use when calling Yarr's ScanConsole

- `-q:` Quiet mode: Reduces the terminal output to two different levels:
  - `-q 0:` No reduction at all
  - `-q 1:` Shows only Yarr's main results + Scan Operator logs  

#### DCS related

- `-o:` **Turns on the output of the channels** specified in `config.json`, after setting V and I to the configuration values and before running the sequence of scans. **Turns the output off again** after running all the scans. If the output is already on, it will just change the voltage / current to the target values.

- `-d:` Monitor the PS and store **real-time DCS data in InfluxDB**.

- `-t:` **After each scan**, retrieves the DCS data from InfluxDB and uploads it to LocalDB, synchronized with each scan. Remember to tell the databases how to talk to each other in the [InfluxDB-LocalDB connectivity file](#influxdb-connectivity-configuration-file). If you want instead to transfer DCS data from InfuxDB to localDB *once all the scans are done*, refer to [this section](#uploading-dcs-to-localdb-offline).

### Example calls

Here are some example use cases of the Scan Operator.

#### *Hello world*

```bash
./ScanOperator.sh -j configs/config.json -m module_id -c
```

This will just run in series the scans you specified in `scan_list` (in `config.json`). All the DCS / LocalDB / InfluxDB functionality is disabled here.

#### *QC procedure*

If you are using the Scan Operator during the QC procedure of any of your modules, you should call it by enabling all the QC / LocalDB Features. (`-Q` and `-W` respectively). You don't need to monitor any DCS data.

```bash
./ScanOperator.sh -j configs/config.json -m module_id -c -Q -W
```

To manually sync DCS in localDB after running this, refer to [the following section](#uploading-dcs-to-localdb-offline).

#### *Monitoring and syncing DCS*

If you want to use the SO to monitor DCS (`-d`) and synchronize the DCS data and the scan data in localDB after each scan (`-t`), use

```bash
./ScanOperator.sh -j configs/m_config.json -m module_id -d configs/m_powersupply.json -W
```

## 4. View the results offline

The SO puts all the scan data under the `data/` folder. There, the data is organized in folders named `YYMMDD_(RUN_COUNTER)`. Each of them contains Yarr's output for all the scans performed under the same call to the SO.


### Monitoring DCS with an external tool

If you are monitoring some DCS variables and you want to associate them with your scan data in localDB, please refer to the folllowig:

- If your tool uploads DCS data to influxDB in real-time, you can always use `-t` to perform the transfer after every single scan.   

- If you prefer to synchronize all the DCS data for all the scans at once, and once the SO has finished running don't use `-t` and you will see a message like this at the end of the sequence of scans:

```
[ info ][so] If you want to upload to localDB the DCS data associated
             to all the scans you have just run (assuming you have it
             stored in InfluxDB), you can do it in one command:
             
                 bash /home/me/scan-operator/data/200923_017/moveDCStoLocalDB.sh path/to/influxdb_connectivity.json
```

That script is automatically generated and contains one call to localdb's dbAccessor per scan ran in the last SO call. The only argument it takes is the influxDB connectivity file we have described [above](#influxdb-connectivity-configuration-file).

# Troubleshooting

These are a few issues you might encounter when using this software.

## 1. How do I know the code of my institution?

You might ask yourself this quesiton if you run into the following error message:

```
[ error  ][   Local DB    ]: Not found site data {'institution': 'localhost.localdomain'} registered in Local DB.
[ error  ][   Local DB    ]: Please set your institution correctly in
[ error  ][   Local DB    ]: { "code": "xxx" } or { "institution": "xxx" } in /home/you/.yarr/localdb/localhost.localdomain_site.json
[ error  ][   Local DB    ]: Invalid configs for uploading data, aborting...
```

### so, how do I know the code of my institution?

1. Go to your Yarr folder
1. Run `./localdb/bin/localdbtool-retrieve list site`

*Note: You might need to switch to devel-localdb to run the above command*


[//]: # (## 6. Other tools included in the package)

[//]: # (###`libDCS/ivscan.py`)

[//]: # (Used to perform sensor IV / module VI scans, requires labRemote.)


## 2. My module is registered in iTk PD, but it doesn't appear in my localDB


The following error will terminate the program if running on `-Q` mode:
```
[ error  ][   Local DB    ]: Not found component data in Local DB: 20UPGM20000002
```

To solve it, you need to have [localdb-tools](https://gitlab.cern.ch/YARR/localdb-tools) installed somewhere on your computer. Then you can retrieve the module from iTkPD to localDB, following this:

1. `cd localdb-tools/viewer/itkpd-interface/`
1. `source authenticate.sh`
2. `./bin/downloader.py --option Module`

See `./bin/downloader.py -h` for more information about this retriever tool.
